package teraUtils;
use strict;

return 1;

sub compareMatrices {
    my ($matrix1, $matrix2,$silent) = @_;

    if( !-e $matrix1 ) {
        return( 0, "could not find first matrix file" );
    }
    if( !-e $matrix2 ) {
        return( 0, "could not find second matrix file" );
    }
    open( MATRIX1, $matrix1 ) or 
        return( 0, "could not open first matrix file" );
    open( MATRIX2, $matrix2 ) or
        return( 0, "could not open second matrix file" );

    my $success = 1;
    my @notes;
    my $lineCnt = 0;
    my $diffCnt = 0;
    my $realDiffCnt = 0;
    my $inMatrix = 0;
    while(<MATRIX1>) {
        my $line2 = <MATRIX2>;
        if( !$line2 ) {
            return( 0, "file2 shorter");
        }
        if( /node/ ) {
            push @notes, "ARGS: " . $lineCnt . " lines, different: " 
                . $diffCnt if $diffCnt;

            $diffCnt = 0;
            $lineCnt = 0;
        }


        if( $line2 ne $_ ) {
            $diffCnt++;
            my @tokens1 = split ',', $_;
            my @tokens2 = split ',', $line2;
            my $realDiff = 0;
            for( my $i=0; $i<@tokens1; $i++ ) {
                if( $inMatrix ) {
                    if( abs($tokens1[$i]-$tokens2[$i]) > 0.00001 ) {
                        $realDiff = 1;
#                    print ", " . $tokens1[$i] . "/" . $tokens2[$i];
my $cladeNum = $i-1;
print "$cladeNum $tokens1[0] $tokens1[$i] $tokens2[$i]$/" if !$silent;
                    }
                } elsif( $tokens1[$i] ne $tokens2[$i] ) {
#                print ", " . $tokens1[$i] . "/" . $tokens2[$i];
                }
            }
            $realDiffCnt++ if $realDiff;
print "$tokens1[0]$/" if $realDiff && !$silent;
#        print "$/";
        }

        $lineCnt++;
        if( /node/ ) {
            #print "header: " . $lineCnt . " lines, different: " . $diffCnt . "$/";
            $diffCnt = 0;
            $lineCnt = 0;
            $inMatrix = 1;
        } elsif( /bestReceiversCost/ ) {
            if( $realDiffCnt ) {
                push @notes, "br: " . $lineCnt . " lines, different: " 
                    . $diffCnt;
                #$success = 0; # good as long as costs and matrix are the same
            }

            $diffCnt = 0;
            $lineCnt = 0;
        } elsif( /bestReceivers/ ) {
            if( $realDiffCnt ) {
                push @notes, "matrix: " . $lineCnt . " lines, different: " 
                    . $realDiffCnt;
                $success = 0;
            }
            $diffCnt = 0;
            $lineCnt = 0;
            $inMatrix = 0;
        }
    }
    if( <MATRIX2> ) {
        return( 0, "file2 longer");
    }
    close MATRIX1;
    close MATRIX2;

    if( $realDiffCnt ) {
        push @notes, "br cost: " . $lineCnt . " lines, different: " 
            . $realDiffCnt;
        $success = 0;
    }

    return ($success, @notes);
}



sub compareGraphs {
    my ($file1, $file2) = @_;

    if( !-e $file1 ) {
        return( 0, "could not find first graph file" );
    }
    if( !-e $file2 ) {
        return( 0, "could not find second graph file" );
    }

    open FILE1, $file1 or die "Could not open $file1";
    open FILE2, $file2 or die "Could not open $file2";

    # line by line comparison
    my @notes;
    my $success = 1;
    while( <FILE1> ) {
        my $line2 = <FILE2>;

        if( $_ eq $line2 ) {
            next;
        }

        my $cladeNum;
        my $spNum;
        my $cost;
        my $event;
        my $support;
        if( /^([A-Z]+_\d+),(\d+) -> \((\d+),(\d+),([0-9.]+)\)\s*$/ ) {
            $cladeNum = $1;
            $spNum = $2;
            $cost = $3;
            $event = $4;
            $support = $5;
        } elsif( /^\((\d+),(\d+),([0-9.]+)\) -> ([A-Z]+_\d+),(\d+)\s*$/ ) {
            $event = $1;
            $support = $2;
            $cladeNum = $3;
            $spNum = $4;
            $cost = $5;
        } else {
            return( 0, "Bad format in first file $_" );
        }

        my $cladeNum2;
        my $spNum2;
        my $cost2;
        my $event2;
        my $support2;
        if( $line2 =~ /^([A-Z]+_\d+),(\d+) -> \((\d+),(\d+),([0-9.]+)\)\s*$/ ) {
            $cladeNum2 = $1;
            $spNum2 = $2;
            $cost2 = $3;
            $event2 = $4;
            $support2 = $5;
        } elsif($line2=~/^\((\d+),(\d+),([0-9.]+)\) -> ([A-Z]+_\d+),(\d+)\s*$/)
        {
            $event2 = $1;
            $support2 = $2;
            $cladeNum2 = $3;
            $spNum2 = $4;
            $cost2 = $5;
        } else {
            return( 0, "Bad format in second file $_" );
        }

        if( $cladeNum != $cladeNum2 && $spNum != $spNum2
            && $cost != $cost2 && $event != $event2
            && abs($support-$support2) > 0.000001 ) {
            return( 0, "No match on $line2 and <$_" );
        }

    }
    close FILE1;
    close FILE2;

    return ($success, @notes);
}

# read the TERA recon format
#55:71,S,15,70@1;70,SL,32@1;69,SL,41@1;68,S,42,67@1:11,54
sub readRecon {
    my ($reconFile) = @_;

    my $root;
    my %eventLists;
    my %geneParents;
    my %geneChildren;
    open RECON, $reconFile or die "Could not open $reconFile";
    while( <RECON> ) {
        chomp;
        next if /^\s*$/;
        if( /^(.+):(.+):(.+),(.+)\s*$/ ) {
            my $geneId = $1;
            my $eventStr = $2;
            my $childL = $3;
            my $childR = $4;
            $eventStr =~ s/'//g;
            my @events = split ';', $eventStr;
            shift @events if $root; #first event is repeated previous last
            $eventLists{$geneId} = \@events;
            push @{$geneChildren{$geneId}}, $childL;
            push @{$geneChildren{$geneId}}, $childR;
            $geneParents{$childL} = $geneId;
            $geneParents{$childR} = $geneId;
            $root = $geneId if !$root;
        } elsif( /^(.+):(.+)\s*$/ ) {
            my $geneId = $1;
            my $eventStr = $2;
            $eventStr =~ s/'//g;
            my @events = split ';', $eventStr;
            shift @events if( $root ); #first event is repeated lasted event
            $eventLists{$geneId} = \@events;
            $root = $geneId if !$root;
        } else {
            die "Bad line: $_";
        }
    }
    close RECON;

    ($root, \%eventLists, \%geneParents, \%geneChildren);
}

my $gInternalIdCounter;
# parse newick tree
sub parseNewick {
    my ($inString, $requireBranchLengths) = @_;

    my $treeString;
    my $rootId;
    if( $inString =~ /^\((.*)\)([^():,]*);\s*$/ ) {
        $treeString = $1;
        $rootId = $2;
    } else {
        die "teraUtils::parseNewick: invalid Newick format";
    }

    if( !$rootId ) {
        $rootId = "*1";
        $gInternalIdCounter = 1;
    }

    my %parent;
    my %children;
    my %branchLengths;
    parseNewickAux( $treeString, $rootId, \%parent, \%children,
                    \%branchLengths, $requireBranchLengths);

    return ($rootId,\%parent,\%children,\%branchLengths);
}

sub parseNewickAux{
    my ($treeString, $parentId, $parentRef, $childrenRef,
        $branchLengthsRef, $requireBranchLengths) = @_;

    while( $treeString ) {
        if( $treeString =~ /^(.*[,(])([^():,]+):([0-9.eE+-]+)$/ 
            || $treeString =~ /^(.*[,(])([^():,]+)$/ ) {
            # leaf
            my $id = $2; #leaf name
            if( $requireBranchLengths && !$3 ) {
                die "teraUtils::parseNewick No branch lengths in $treeString";
            }
            $branchLengthsRef->{$id} = $3;
            $parentRef->{$id} = $parentId;
            unshift @{$childrenRef->{$parentId}}, $id;
            $treeString = $1
        } elsif( $treeString =~ /^(.*),$/ ) {
            # sibling
            $treeString = $1
        } elsif( $treeString =~ /^(.*)\)([^():,]*):([0-9.eE+-]+)$/ 
                 || $treeString =~ /^(.*)\)([^():,]*)$/ ) {
            # internal node
            my $id = $2;
            if( !$id ) {
                $gInternalIdCounter += 1;
                $id = "*$gInternalIdCounter";
            }
            if( $requireBranchLengths && !$3 ) {
                die "teraUtils::parseNewick No branch lengths in $treeString";
            }
            $branchLengthsRef->{$id} = $3;
            $parentRef->{$id} = $parentId;
            unshift @{$childrenRef->{$parentId}}, $id;
            $treeString = parseNewickAux( $1, $id, $parentRef, $childrenRef, 
                                          $branchLengthsRef, 
                                          $requireBranchLengths );
        } elsif( $treeString =~ /^(.*)\($/ ) {
            # end of internal node
            return $1;
        } else {
            die "teraUtils::parseNewick Bad tree format: $treeString";
        }
    }
}

